#!/bin/bash
#
# This build script compiles a GCC based cross compiler using the
# newlib C library.
# Known to work under GNU/Linux and Cygwin targeting m68k-elf
# 

srcdir=`pwd`

#
#  Prints the usage for this script.
#
usage()
{
  echo "Usage:"
  echo "$0 installdir [srcdir]"
  echo
  echo "srcdir contains the gzipped, tarred packages and patches. By default this"
  echo "is taken to be the path with which this script was run."
}

#
#  Checks the status returned by executables and exits if it is non-zero.
#
check_fatal()
{
  if [ $1 -ne 0 ] ; then
    shift
    echo "ERROR: $*" >&2
    exit 1
  fi
}

# target platform
TARGET=m68k-elf

# Base packages
BINUTILS=binutils-2.10
GCC=gcc-2.95.3
NEWLIB=newlib-1.9.0

# Patches we apply to the base packages
BINUTILS_PATCHES="binutils-2.10-coldfire-fixes.diff binutils-2.10-coldfire-cpushl.diff"
GCC_PATCHES="gcc-2.95.3-coldfire-epilogue-fix.diff gcc-2.95.3-coldfire-hwdiv.diff gcc-2.95.3-coldfire-longmul.diff \
gcc-2.95.3-interrupt-attribute-20010127.diff gcc-2.95.3-coldfire-version-message.diff gcc-2.95.3-coldfire-alignment.diff"
NEWLIB_PATCHES="newlib-1.9.0-coldfire-20010318.diff"

# adjust these if you download tar.gz archives. All archives must use the same compressor. Sorry.
COMPRESSOR=bzip2
ARCHIVE_EXT=tar.bz2

#set this variable if you want a set of compiler tarballs for a ColdFire/m68k cross-compiler
COLDFIRE_GCC_BUILD="yes"
COLDFIRE_GCC_VERSION="r7-20010415"

#determine the OS
case `uname -s` in
  Linux)   BUILD_OS=linux ; EXE_EXT="" ; README="README" ;;
  CYGWIN*) BUILD_OS=win32 ; EXE_EXT=".exe" README="readme.txt" ;;
  *)       BUILD_OS=unknown ; EXE_EXT="" ; README="README" ;;
esac

#display script information
echo "build-gcc.sh - GCC cross-compiler build script"
echo

# check that we have a valid set of parameters
if [ $# -ne 1 ]; then
  usage
  exit 1
fi

dstdir="$1"

if [ -n "$2" ]; then
    srcdir="$2"
fi


#
#  If the specified tools are not available, then abort.
#
echo "Checking for tool source in ${srcdir}"
test -f ${srcdir}/${BINUTILS}.${ARCHIVE_EXT} || check_fatal 1 "Cannot find BINUTILS (${BINUTILS}.${ARCHIVE_EXT})"
test -f ${srcdir}/${GCC}.${ARCHIVE_EXT}      || check_fatal 1 "Cannot find GCC (${GCC}.${ARCHIVE_EXT})"
test -f ${srcdir}/${NEWLIB}.${ARCHIVE_EXT}   || check_fatal 1 "Cannot find NEWLIB (${NEWLIB}.${ARCHIVE_EXT})"

# check the patches all exist too
for patch_file in ${BINUTILS_PATCHES} ${GCC_PATCHES} ${NEWLIB_PATCHES}; do
  test -f ${srcdir}/${patch_file} || check_fatal 1 "Cannot find patch file ${patch_file}"
done

if [ "x${COLDFIRE_GCC_BUILD}" = "xyes" ] ; then
  test -d ${srcdir}/examples  || check_fatal 1 "Cannot find the ColdFire examples"
  test -f ${srcdir}/${README}  || check_fatal 1 "Cannot find the ColdFire ${README}"
fi
    
echo "
Building ${TARGET} cross toolset under ${BUILD_OS} and installing in ${dstdir}
"

# cleanup any existing source directory and create a new one
if [ -d source ]; then
  echo "Removing old source directory"
  rm -Rf source
  echo "done"
fi

mkdir source
cd source

# extract and patch each toolset
echo "Extracting binutils from archive..."
${COMPRESSOR} -d -c ${srcdir}/${BINUTILS}.${ARCHIVE_EXT} | tar -x
echo "patching..."
for patch_file in ${BINUTILS_PATCHES}; do
  (cd ${BINUTILS}; patch -p1 < ${srcdir}/${patch_file})
done
echo "done"

echo "Extracting gcc from archive..."
${COMPRESSOR} -d -c ${srcdir}/${GCC}.${ARCHIVE_EXT} | tar -x
echo "patching..."
for patch_file in ${GCC_PATCHES}; do
  (cd ${GCC}; patch -p1 < ${srcdir}/${patch_file})
done
echo "done"

echo "Extracting newlib from archive..."
${COMPRESSOR} -d -c ${srcdir}/${NEWLIB}.${ARCHIVE_EXT} | tar -x
echo "patching..."
for patch_file in ${NEWLIB_PATCHES}; do
  (cd ${NEWLIB}; patch -p1 < ${srcdir}/${patch_file})
done
echo "done"

echo "Removing unneeded languages (Java, chill, fortran, Objective C) and linking newlib to gcc"
(
  cd ${GCC}
  
  # delete the unused languages. Much more reliable than mucking with configure or LANGUAGES env var
  rm -Rf gcc/f
  rm -Rf gcc/ch
  rm -Rf gcc/objc
  rm -Rf gcc/java
  rm -Rf libchill
  rm -Rf libf2c
  rm -Rf libobjc

  # symlink newlib and libgloss into gcc
  ln -s ../${NEWLIB}/newlib .  
  ln -s ../${NEWLIB}/libgloss .
)
echo "done"

echo "Create a build directory"
cd ..
if [ -d build ]; then
  echo "Removing old build directory"
  rm -Rf build
  echo "done"
fi

mkdir build
cd build
echo "done"

# Build and install binutils separately becuase a one-tree build doesn't work under Cygwin
mkdir binutils
(
  cd binutils
  echo "Configuring binutils (for details see build/binutils-configure.log)"
  ../../source/${BINUTILS}/configure --target=${TARGET} --prefix=${dstdir} > ../binutils-configure.log 2>&1

  echo "Building binutils (for details see build/binutils-make.log)"
  make > ../binutils-make.log 2>&1

  echo "Installing binutils (for details see build/binutils-install.log)"
  make install > ../binutils-install.log 2>&1

  echo "done"
)

echo "Adding binutils to our PATH"
export PATH=$PATH:${dstdir}/bin
echo "PATH=$PATH"


# Once we have a working binutils for our target we can build gcc and newlib together
mkdir gcc-newlib
(
  cd gcc-newlib
  echo "Configuring gcc and newlib (for details see build/gcc-newlib-configure.log)"
  ../../source/${GCC}/configure --target=${TARGET} --prefix=${dstdir} --with-newlib > ../gcc-newlib-configure.log 2>&1

  echo "Building gcc and newlib (for details see build/gcc-newlib-make.log)"
  make > ../gcc-newlib-make.log 2>&1

  echo "Installing gcc and newlib (for details see build/gcc-newlib-install.log)"
  make install > ../gcc-newlib-install.log 2>&1

  echo "done"
)

if [ "x${COLDFIRE_GCC_BUILD}" = "xyes" ] ; then
  echo "striping unneeded debug information from binaries"
  strip ${dstdir}/bin/*
  strip ${dstdir}/${TARGET}/bin/*
  strip ${dstdir}/lib/gcc-lib/m68k-elf/2.95.3/cc1${EXE_EXT}
  strip ${dstdir}/lib/gcc-lib/m68k-elf/2.95.3/cc1plus${EXE_EXT}
  strip ${dstdir}/lib/gcc-lib/m68k-elf/2.95.3/cpp${EXE_EXT}
  strip ${dstdir}/lib/gcc-lib/m68k-elf/2.95.3/collect2${EXE_EXT}
  echo "done"

  echo "deleting naff documentation formats and unused libraries"
  rm -rf ${dstdir}/info
  rm -rf ${dstdir}/man
  rm -rf ${dstdir}/share
  rm ${dstdir}/lib/*
  rm ${dstdir}/include/*
  echo "done"

  echo "copying ColdFire examples"
  cp -a ${srcdir}/examples ${dstdir}
  echo "done"
  
  echo "creating gcc-m68k, extralibs and cpu32libs"
  if [ -d package ]; then
    echo "Removing old package directory"
    rm -Rf package
    echo "done"
  fi
  
  mkdir package
  (
    cd package
  
    #first we copy the pristine installed tree into 3 directories for each package
    mkdir -p gcc-m68k-${BUILD_OS}-${COLDFIRE_GCC_VERSION}/${dstdir}
    mkdir -p cpu32libs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}/${dstdir}
    mkdir -p extralibs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}/${dstdir}
    cp -a ${dstdir}/* gcc-m68k-${BUILD_OS}-${COLDFIRE_GCC_VERSION}/${dstdir}
    cp -a ${dstdir}/* cpu32libs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}/${dstdir}
    cp -a ${dstdir}/* extralibs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}/${dstdir}
  
    #cleanup gcc-m68k to only have coldfire libraries
    (
      cd gcc-m68k-${BUILD_OS}-${COLDFIRE_GCC_VERSION}/${dstdir}
      cp -a ${srcdir}/${README} .
      rm lib/gcc-lib/m68k-elf/2.95.3/libgcc.a
      rm lib/gcc-lib/m68k-elf/2.95.3/libstdc++.a
      rm -r lib/gcc-lib/m68k-elf/2.95.3/mcpu32
      rm -r lib/gcc-lib/m68k-elf/2.95.3/m68000
      rm -r lib/gcc-lib/m68k-elf/2.95.3/msoft-float
      rm m68k-elf/lib/*
      rm -r m68k-elf/lib/ldscripts
      rm -r m68k-elf/lib/mcpu32
      rm -r m68k-elf/lib/m68000
      rm -r m68k-elf/lib/msoft-float
      rm m68k-elf/lib/m5200/libmvme*.a
      rm m68k-elf/lib/m5200/mvme*.ld
      rm m68k-elf/lib/m5200/libbcc.a
      rm m68k-elf/lib/m5200/bcc.ld
      rm m68k-elf/lib/m5200/idp.ld
      rm m68k-elf/lib/m5200/libidp.a
      rm m68k-elf/lib/m5200/mhwdiv/libmvme*.a
      rm m68k-elf/lib/m5200/mhwdiv/mvme*.ld
      rm m68k-elf/lib/m5200/mhwdiv/libbcc.a
      rm m68k-elf/lib/m5200/mhwdiv/bcc.ld
      rm m68k-elf/lib/m5200/mhwdiv/idp.ld
      rm m68k-elf/lib/m5200/mhwdiv/libidp.a
      rm m68k-elf/lib/m5200/mhwdiv/sbc5204.ld
      rm m68k-elf/lib/m5200/mhwdiv/sbc5206.ld
    )

    #cleanup cpu32libs to only have CPU32 libraries and no compiler
    (
      cd cpu32libs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}/${dstdir}
      rm -r bin
      rm -r include
      rm -r examples
      rm lib/gcc-lib/m68k-elf/2.95.3/*
      rm -r lib/gcc-lib/m68k-elf/2.95.3/m68000
      rm -r lib/gcc-lib/m68k-elf/2.95.3/msoft-float
      rm -r lib/gcc-lib/m68k-elf/2.95.3/m5200
      rm -r lib/gcc-lib/m68k-elf/2.95.3/include
      rm -r m68k-elf/bin
      rm -r m68k-elf/include
      rm m68k-elf/lib/*
      rm -r m68k-elf/lib/ldscripts
      rm -r m68k-elf/lib/m68000
      rm -r m68k-elf/lib/msoft-float
      rm -r m68k-elf/lib/m5200
      rm m68k-elf/lib/mcpu32/sbc*
      rm m68k-elf/lib/mcpu32/libdbug.a
    )
    
    #cleanup extralibs to only have the remaining missing libraries and no compiler
    (
      cd extralibs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}/${dstdir}
      rm -r bin
      rm -r include
      rm -r examples
      rm lib/gcc-lib/m68k-elf/2.95.3/c*
      rm lib/gcc-lib/m68k-elf/2.95.3/SYSCALLS.c.X
      rm lib/gcc-lib/m68k-elf/2.95.3/specs
      rm -r lib/gcc-lib/m68k-elf/2.95.3/include
      rm -r lib/gcc-lib/m68k-elf/2.95.3/m5200
      rm -r lib/gcc-lib/m68k-elf/2.95.3/mcpu32
      rm -r m68k-elf/bin
      rm -r m68k-elf/include
      find m68k-elf/lib -name "sbc*" -exec rm {} \;
      find m68k-elf/lib -name "libdbug.a" -exec rm {} \;
      rm -r m68k-elf/lib/m5200
      rm -r m68k-elf/lib/mcpu32
    )
    echo "creating gcc-m68k-${BUILD_OS}-${COLDFIRE_GCC_VERSION}"
    (
      cd gcc-m68k-${BUILD_OS}-${COLDFIRE_GCC_VERSION}
      tar -cf ../gcc-m68k-${BUILD_OS}-${COLDFIRE_GCC_VERSION}.tar.bz2 . --use-compress-program=bzip2
    )
    echo "creating cpu32libs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}"
    (
      cd cpu32libs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}
      tar -cf ../cpu32libs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}.tar.bz2 . --use-compress-program=bzip2
    )
    echo "creating extralibs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}"
    (
      cd extralibs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}
      tar -cf ../extralibs-${BUILD_OS}-${COLDFIRE_GCC_VERSION}.tar.bz2 . --use-compress-program=bzip2
    )
  )
  echo "done"
fi

